package main

import (
	"flag"
	"fmt"
	"os/exec"
	"path/filepath"
	"runtime"
	"strconv"
	"sync"
	"time"

	"git.milar.in/milarin/channel"
)

var ModulePath string
var ProjectName string

var (
	OS   = flag.String("os", "", "comma-separated list of operating systems to compile for (empty for all)")
	Arch = flag.String("arch", "", "comma-separated list of architectures to compile for (empty for all)")

	Silent     = flag.Bool("s", false, "silent mode (no output)")
	NoCompress = flag.Bool("c", false, "dont compress any executables")
	NumThreads = flag.Int("t", runtime.NumCPU(), "amount of threads (0 = infinite)")
)

var Runner channel.Runner

func main() {
	var err error
	flag.Parse()

	if err := FillCompileConfigs(); err != nil {
		panic(err)
	}

	ModulePath, err = filepath.Abs(flag.Arg(0))
	if err != nil {
		panic(err)
	}
	ProjectName = filepath.Base(ModulePath)

	ch := make(chan *CompileReport, len(CompileConfigs))
	wg := new(sync.WaitGroup)

	doneCh := make(chan struct{})
	go Watch(ch, doneCh)

	Runner = GetRunner()

	start := time.Now()

	for _, cfg := range CompileConfigs {
		cfg := cfg
		wg.Add(1)
		go Runner.Run(func() { Compile(cfg, ch, wg) })
	}

	wg.Wait()
	close(ch)

	end := time.Now()

	<-doneCh // wait for Watch

	if !*Silent {
		fmt.Printf("compilation took %s (using %s threads)\n", end.Sub(start), GetThreadCount())
	}
}

func Compile(cfg *CompileConfig, ch chan<- *CompileReport, wg *sync.WaitGroup) {
	filename := fmt.Sprintf("%s_%s_%s%s", ProjectName, cfg.OS, cfg.Arch, cfg.FileExt)

	ch <- &CompileReport{Config: cfg, State: StateCompiling}

	compileCmd := exec.Command("go", "build", "-o", filename)
	compileCmd.Dir = ModulePath
	if err := compileCmd.Start(); err != nil {
		ch <- &CompileReport{Config: cfg, State: StateCompileError}
		wg.Done()
		return
	}
	if err := compileCmd.Wait(); err != nil {
		ch <- &CompileReport{Config: cfg, State: StateCompileError}
		wg.Done()
		return
	}

	Compress(filename, cfg, ch, wg)

	// uncomment for independent compile and compress tasks
	/*
		ch <- &CompileReport{Config: cfg, State: StateWaiting}
		go Runner.Run(func() { Compress(filename, cfg, ch, wg) })
	*/
}

func Compress(filename string, cfg *CompileConfig, ch chan<- *CompileReport, wg *sync.WaitGroup) {
	defer wg.Done()

	ch <- &CompileReport{Config: cfg, State: StateCompressing}

	if !*NoCompress && cfg.Compress {
		compressCmd := exec.Command("upx", "--best", filename)
		compressCmd.Dir = ModulePath

		if err := compressCmd.Start(); err != nil {
			ch <- &CompileReport{Config: cfg, State: StateCompressError}
			return
		}

		if err := compressCmd.Wait(); err != nil {
			ch <- &CompileReport{Config: cfg, State: StateCompressError}
			return
		}
	}

	ch <- &CompileReport{Config: cfg, State: StateDone}
}

func GetRunner() channel.Runner {
	if *NumThreads <= 0 {
		return channel.NewUnlimitedRunner()
	} else {
		return channel.NewLimitedRunner(*NumThreads)
	}
}

func GetThreadCount() string {
	if *NumThreads == 0 {
		return "infinitely many"
	}
	return strconv.Itoa(*NumThreads)
}
